/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:fr.bigeon.gclc.command.ScriptExecution.java
 * Created on: Jun 12, 2016
 */
package fr.bigeon.gclc.command;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.List;

import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.GCLCConstants;
import fr.bigeon.gclc.exception.CommandParsingException;
import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.CommandRunExceptionType;

/** A command that will laucnh a series of command from a file
 * <p>
 * This command will read a file and execute each non empty non commented line
 * as a command of the application.
 *
 * @author Emmanuel Bigeon */
public class ScriptExecution extends Command {

    /** The tab character */
    private static final String TAB = "\t"; //$NON-NLS-1$
    /** the space character */
    private static final String SPACE = " "; //$NON-NLS-1$
    /** The application */
    private final ConsoleApplication application;
    /** The commenting prefix */
    private final String commentPrefix;
    /** The charset for files */
    private final Charset charset;

    /** @param name the name of the command
     * @param application the application
     * @param commentPrefix the comment prefix in the script files
     * @param charset the charset to use for files */
    public ScriptExecution(String name, ConsoleApplication application,
            String commentPrefix, Charset charset) {
        super(name);
        this.application = application;
        this.commentPrefix = commentPrefix;
        this.charset = charset;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommand#execute(java.lang.String[]) */
    @Override
    public void execute(String... args) throws CommandRunException {
        checkArgs(args);
        String scriptFile = args[0];
        String[] params = Arrays.copyOfRange(args, 1, args.length);
        String cmd;
        int lineNo = -1;
        try (InputStreamReader fReader = new InputStreamReader(
                new FileInputStream(scriptFile), charset);
             BufferedReader reader = new BufferedReader(fReader)) {
            while ((cmd = reader.readLine()) != null) {
                lineNo++;
                String cmdLine = readCommandLine(cmd, params);
                if (cmdLine == null) {
                    continue;
                }
                List<String> ps = GCLCConstants.splitCommand(cmdLine);
                String command = ps.remove(0);
                application.executeSub(command, ps.toArray(new String[0]));
            }
        } catch (CommandParsingException e) {
            throw new CommandRunException(MessageFormat.format(
                    "Invalid command in script ({0})", e.getLocalizedMessage()), //$NON-NLS-1$
                    e, this);
        } catch (IOException e) {
            throw new CommandRunException("Unable to read script", //$NON-NLS-1$
                    e, this);
        } catch (CommandRunException e) {
            throw manageRunException(e, lineNo);
        }
    }

    /** @param args the arguments
     * @throws CommandRunException if the arguments were not the ones
     *             expected */
    private void checkArgs(String[] args) throws CommandRunException {
        if (args.length == 0) {
            throw new CommandRunException(CommandRunExceptionType.USAGE,
                    "Expecting a file", this); //$NON-NLS-1$
        }
    }

    /** This method will create the correct exception. The exception source must
     * be this command.
     * 
     * @param e the exception
     * @param lineNo the line nu;ber
     * @return the exception to actually throw */
    private CommandRunException manageRunException(CommandRunException e,
                                                   int lineNo) {
        if (e.getSource() == this) {
            // ensure closing?
            return e;
        }
        return new CommandRunException(CommandRunExceptionType.EXECUTION,
                MessageFormat.format(
                        "The script could not complete due to command failure at line {0} ({1})", //$NON-NLS-1$
                        Integer.valueOf(lineNo), e.getLocalizedMessage()),
                e, this);
    }

    /** @param cmd the line
     * @param params the formatting parameters
     * @return the command if it is indeed one, null otherwise
     * @throws CommandRunException if the line stqrted with a space character */
    private String readCommandLine(String cmd,
                                   Object[] params) throws CommandRunException {
        if (cmd.startsWith(SPACE) || cmd.startsWith(TAB)) {
            throw new CommandRunException(
                    "Invalid command in script (line starts with space character)", //$NON-NLS-1$
                    this);
        }
        if (cmd.isEmpty() || cmd.startsWith(commentPrefix)) {
            return null;
        }
        return MessageFormat.format(cmd, params);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.Command#usageDetail() */
    @SuppressWarnings("nls")
    @Override
    protected String usageDetail() {
        StringBuilder builder = new StringBuilder();
        builder.append(
                "  scriptfile: path to the file containing the script to execute.");
        builder.append(System.lineSeparator());
        builder.append(System.lineSeparator());
        builder.append(
                "  The script file must contain one line commands. The lines must never");
        builder.append(System.lineSeparator());
        builder.append(
                "start with whitespace characters. The lines starting with");
        builder.append(System.lineSeparator());
        builder.append("\"" + commentPrefix +
                       "\" will be ignored as well as empty lines.");
        builder.append(System.lineSeparator());

        return builder.toString();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.Command#usagePattern() */
    @SuppressWarnings("nls")
    @Override
    protected String usagePattern() {
        return super.usagePattern() + " <scriptfile>";
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommand#tip() */
    @SuppressWarnings("nls")
    @Override
    public String tip() {
        return "Execute a script";
    }

}
